/********************************************************************
 *		File			:		iservice.h
 *		Description		:		Com Service Registeration interface
 *		Author			:		Jianwen Zhu
 *		Reference		:		http://www.microsoft.com/com
 *		                 		oskit
 *		Email			:		jzhu@eecg.toronto.edu
 *********************************************************************/
#ifndef ISERVICE_H
#define ISERVICE_H

#include "com.h"

#define COM_ISERVICE_IID  COM_DEFINE_GUID(				\
		0x38b8a7ce, 0xd4ae, 0x11d3, 					\
		0xb0, 0x07, 0x00, 0xb0, 0xd0, 0x19, 0xb1, 0x55	\
		)

COM_DECLARE_INTERFACE( Service );

struct _DService {
	COMDECL			(*query)( IService si, const ComIID iid, IGeneric *i );
	COMDECLU		(*addref)( IService si);
	COMDECLU		(*release)( IService si);

	COMDECLS		(*dump)( IService si, FILE* fp, char* bp );
	COMDECLP		(*create)( IService si, IMem mem );
	COMDECL			(*destruct)( IService si );	
	COMDECL			(*construct)( IService si );
	
	COMDECL			(*constructWithMem)( IService si, IMem mem );
	COMDECL			(*add)( IService si, const ComIID iid, IGeneric i );
	COMDECL			(*remove)( IService si, const ComIID iid, IGeneric i );
	COMDECL			(*lookup)(
		IService si, const ComIID iid, IGeneric ** iarray, IMem mem
		);
	COMDECL			(*lookupFirst)(
		IService si, const ComIID iid, IGeneric *iref
		);
	COMDECL			(*clone)( IService si, IService* another );
	};

#ifdef __cplusplus
class CService : public CComponent {
	public:

	CService( IGeneric i ) : CComponent( i ) {}
	CService( void ) {}
	
	inline Jerror construct( CMem mem ) {
		return COM_CALL( constructWithMem, (IService)intf )(
			(IService)intf, (IMem)mem.intf
			);
		}
	inline Jerror add( const ComIID iid, CUnKnown i ) {
		return COM_CALL( add, (IService)intf )(
			(IService)intf, iid, (IGeneric)i.intf
			);
		}
	inline Jerror remove( const ComIID iid, CUnKnown i ) {
		return COM_CALL( add, (IService)intf )(
			(IService)intf, iid, (IGeneric)i.intf
			);
		}
	inline Jerror lookup( const ComIID iid, IGeneric** iarray, CMem mem ) {
		return COM_CALL( lookup, (IService)intf )(
			(IService)intf, iid, iarray, (IMem)mem.intf
			);
		}
	inline Jerror lookupFirst( const ComIID iid, CUnKnown& i ) {
		IUnKnown	found;
		Jerror 		rc = COM_CALL( lookupFirst, (IService)intf )(
			(IService)intf, iid, (IGeneric *)&found
			);
		i.set( found );
		return rc;
		}
	inline Jerror clone( const ComIID iid, CService& another ) {
		IUnKnown	found;
		Jerror 		rc = COM_CALL( clone, (IService)intf )(
			(IService)intf, (IService *)&found
			);
		another.set( found );
		return rc;
		}
	};
#endif

extern IService		iservice;

#endif
