/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/* The purpose of this file is to provide fast 
   DdNode Pointer to BddClass lookup.
   bcl = BddClassLookup */

#include <assert.h>

#include "metautil.h"
#include "fbddutil.h"

struct _BclManager
{
  struct _HashTable hashTable;
};

BclManager
bclNewManager()
{
  BclManager bclm = 
    (BclManager)comAlloc(_fbddHeap, sizeof(struct _BclManager));
  hashDefaultConstruct(&bclm->hashTable, _fbddHeap);
  return bclm;
}

void
bclDelManager(BclManager bclm)
{
  hashDestruct(&bclm->hashTable);
  comFree(_fbddHeap, bclm);
}

BddClass *
bclGet(BclManager bclm, DdNode *node)
{
  return (BddClass*)(hashGet(&bclm->hashTable, node));
}

void
bclPut(BclManager bclm, DdNode *node, BddClass *bddClass)
{
  hashPut(&bclm->hashTable, (void*)node, (void*)bddClass);
}

void
bclRemove(BclManager bclm, DdNode *node)
{
  hashDel(&bclm->hashTable, (void*)node);
}

void
bclRemoveBddi(BclManager bclm, Model *model, BddInstance *bddi)
{

  listDelElement(&bddi->bddClass->lInsts, bddi->cBddcInst);

  if (bddi->bddClass->type == PRIMARY_OUTPUT)
    return;

  if (bddcNumInsts(bddi->bddClass) == 0
      && isBasicGate(model, bddi->bddClass)==FALSE){
    if (bclGet(bclm, bddi->bddClass->bdd)){
      bclRemove(bclm, bddi->bddClass->bdd);
    }
  }
}

// check each entry in bcl is reachable.
int
bclCheck(BclManager bclm, Model *model)
{
  struct _List lBddcs;
  int i, j;
  HashCell hc;
  HashTable hashTable = &bclm->hashTable;

  listConstruct0(&lBddcs, _fbddHeap);
  modelBddcList(_fbddHeap, model, &lBddcs);
  
  hashForeach(hashTable, i, j, hc){
    BddClass *bddcMatch = hashData(BddClass*, hc);
    char found = FALSE;

    if (!isBasicGate(model, bddcMatch)){
      ListCell cBddc;
      listForeach(&lBddcs, cBddc){
	BddClass *bddc = listData(BddClass*, cBddc);
	if (bddc == bddcMatch){
	  found = TRUE;
	  break;
	}
      }
      assert(found);
    }
  }
  listDelAll(&lBddcs);

  return TRUE;
}


void
populateBclWithBasicGates(BclManager bclm, Model *model)
{
  bclPut(bclm, model->nodeLogicZero, model->bddcLogicZero);
  bclPut(bclm, model->nodeLogicOne, model->bddcLogicOne);
  bclPut(bclm, model->nodeSF, model->bddcSF);
  bclPut(bclm, model->nodeMUX, model->bddcMux);
  bclPut(bclm, model->nodeMUX2, model->bddcMux2);
  bclPut(bclm, model->nodeAND, model->bddcAnd);
  bclPut(bclm, model->nodeOR, model->bddcOr);
  bclPut(bclm, model->nodeXNOR, model->bddcXnor);
  bclPut(bclm, model->nodeNOT, model->bddcNot);
  bclPut(bclm, model->nodeNAND, model->bddcNand);
  bclPut(bclm, model->nodeNOR, model->bddcNor);
  bclPut(bclm, model->nodeXOR, model->bddcXor);
}
