/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

 /* blif parser.  Combinational Logic only.  No support for subcircuits */

%{
  #include "stdlib.h"
  #include "metautil.h"
  #include "reader.h"

  int yyerror (const char *s);
  int yylex();

  static List _curList; // used when creating name lists
  static IMem _arena;
  static char modelName[50];

  static char str1[50], str2[50], str3[50], str4[50], str5[50];
%}
%union {
  char *str;
  int val;
  void *list;
}

%token <val> MODEL_CMD
%token <val> INPUT_CMD
%token <val> OUTPUT_CMD
%token <val> CLOCK_CMD
%token <val> DEFAULT_INPUT_ARRIVAL_CMD
%token <val> DEFAULT_OUTPUT_REQUIRED_CMD
%token <val> DEFAULT_INPUT_DRIVE_CMD
%token <val> DEFAULT_OUTPUT_LOAD_CMD
%token <val> DEFAULT_MAX_INPUT_LOAD_CMD
%token <val> LATCH_CMD
%token <val> NAMES_CMD
%token <val> END_CMD
%token <str> CUBE
%token <str> LATCH_TYPE
%token <str> LATCH_VAL
%token <str> ID
%token <str> FLOAT
%token <val> NEW_LINE
%type <list> input
%type <list> output
%type <list> clock
%type <val> latchVal
%type <str> model_name

%% 
model: 
MODEL_CMD 
{
  _arena = _fbddHeap;
} 
model_name 
input
output 
clock
{
  ListCell cell;

  //  printf("begin_model\n");
  blifModelBegin($3, $4, $5, $6);
  
  listForeach((List)$4, cell){
    comFree(_arena, listData(char*, cell));
  }
  listForeach((List)$5, cell){
    comFree(_arena, listData(char*, cell));
  }
  listDelAll((List)$4);
  listDelAll((List)$5);
  if ((List)$6 != NULL){
    listForeach((List)$6, cell){
      comFree(_arena, listData(char*, cell));
    }
    listDelAll((List)$6);
  comFree(_arena, $6);
  }
  comFree(_arena, $4);
  comFree(_arena, $5);
} 
commands 
END_CMD
NEW_LINE
{
  blifModelEnd();
}
;

model_name:
ID{
  strcpy(modelName, $1);
}
NEW_LINE{
  $$ = modelName;
}
;

input: 
INPUT_CMD 
{
  _curList = comAlloc(_arena, sizeof(struct _List));
  listConstruct0(_curList, _arena);
}
name_list 
{ 
  $$=_curList; 
}
;

output: 
OUTPUT_CMD 
{
  _curList = comAlloc(_arena, sizeof(struct _List));
  listConstruct0( _curList, _arena );
}
name_list 
{ 
  $$=_curList; 
}
;

clock: {$$ = NULL;}
| CLOCK_CMD 
{
  _curList = comAlloc(_arena, sizeof(struct _List));
  listConstruct0( _curList, _arena );
}
name_list 
{ 
  $$=_curList; 
}
;

name_list: 
name_list2
NEW_LINE
;


name_list2: 
ID {
  ListCell cell;
  char *id = comAlloc(_arena, sizeof(char)*(strlen($1)+1));
  strcpy(id, $1);
  listNewCell(_arena, cell, id);

  listAppendElement(_curList, cell);
}
| name_list2 ID {
  ListCell cell;
  char *id = comAlloc(_arena, sizeof(char)*(strlen($2)+1));
  strcpy(id, $2);
  listNewCell(_arena, cell, id);

  listAppendElement(_curList, cell);
}
;

commands: command
  | commands command
;

command: logic_gate {;}
| latch {;}
| default_input_arrival {;}
| default_output_required {;}
| default_input_drive {;}
| default_output_load {;}
| default_max_input_load {;}
;

default_input_arrival: DEFAULT_INPUT_ARRIVAL_CMD FLOAT FLOAT NEW_LINE{
;};

default_output_required: DEFAULT_OUTPUT_REQUIRED_CMD FLOAT FLOAT NEW_LINE{
;};

default_input_drive: DEFAULT_INPUT_DRIVE_CMD FLOAT FLOAT NEW_LINE{
;};

default_output_load: DEFAULT_OUTPUT_LOAD_CMD FLOAT NEW_LINE{
;};

default_max_input_load: DEFAULT_MAX_INPUT_LOAD_CMD FLOAT NEW_LINE{
;};

logic_gate: NAMES_CMD {
  _curList = comAlloc(_arena, sizeof(struct _List));
  listConstruct0(_curList, _arena);
}
name_list { 
  blifLogicGateBegin();
}
single_output_cover { 
  ListCell cell;

  blifLogicGateEnd(_curList);

  listForeach(_curList, cell){
    comFree(_arena, listData(char*, cell));
  }
  listDelAll(_curList);
  comFree(_arena, _curList);
}
;

latch: LATCH_CMD 
ID {strcpy(str1, $2);}
ID {strcpy(str2, $4);}
latchType
latchVal {
    blifLatch(str1, str2, str3, str4, $7);
}
NEW_LINE 
;

latchType: {str3[0]=0; str4[0]=0;}
| LATCH_TYPE {strcpy(str3, $1);} 
ID  {strcpy(str4, $3);}
;

latchVal: {$$=0;}
| LATCH_VAL {$$=atoi($1);}
;

single_output_cover: { blifConstZero(); }
| cubes {;}
;

cubes: CUBE { 
  //  printf("add_cube(%s)\n", $1); 
  blifAddCube($1);
}
NEW_LINE
| cubes CUBE { 
  //  printf("add_cube(%s)\n", $2); 
  blifAddCube($2);
}
NEW_LINE
;

%%

#include <stdio.h>

int
yyerror (const char *s)
{
  printf("%s\n", s);
  return 0;
}
