/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

%{
#include "blifParse.tab.h"
char latchCommand;
int numIDs;
%}

%s CUBE_SEARCH
%s LATCH_OPTIONALS

MODEL_CMD ".model"
INPUT_CMD ".inputs"
OUTPUT_CMD ".outputs"
CLOCK_CMD ".clock"
LATCH_CMD ".latch"
NAMES_CMD ".names"
END_CMD ".end"
DEFAULT_INPUT_ARRIVAL_CMD ".default_input_arrival"
DEFAULT_OUTPUT_REQUIRED_CMD ".default_output_required"
DEFAULT_INPUT_DRIVE_CMD ".default_input_drive"
DEFAULT_OUTPUT_LOAD_CMD ".default_output_load"
DEFAULT_MAX_INPUT_LOAD_CMD ".default_max_input_load"
LATCH_TYPE fe|re|ah|al|as
LATCH_VAL 0|1|2|3
ID [a-zA-Z0-9\[\]][a-zA-Z0-9\[\]\(\)./_\-\<\>$]*
CUBE ([10-]+"\\\n"?" "[10]|1)
NEW_LINE "\n"[ \t\n]*
FLOAT [0-9]*[.][0-9]*

%%

{MODEL_CMD} return MODEL_CMD;
{INPUT_CMD} {
  BEGIN(0);  /* don't mistake ID for a cube */
  return INPUT_CMD;
}
{OUTPUT_CMD} {
  BEGIN(0);  /* don't mistake ID for a cube */
  return OUTPUT_CMD;
}
{CLOCK_CMD} {
  BEGIN(0);
  return CLOCK_CMD;
}
{DEFAULT_INPUT_ARRIVAL_CMD} return DEFAULT_INPUT_ARRIVAL_CMD;
{DEFAULT_OUTPUT_REQUIRED_CMD} return DEFAULT_OUTPUT_REQUIRED_CMD;
{DEFAULT_INPUT_DRIVE_CMD} return DEFAULT_INPUT_DRIVE_CMD;
{DEFAULT_OUTPUT_LOAD_CMD} return DEFAULT_OUTPUT_LOAD_CMD;
{DEFAULT_MAX_INPUT_LOAD_CMD} return DEFAULT_MAX_INPUT_LOAD_CMD;
 
{LATCH_CMD} {
  BEGIN(0);  /* don't mistake ID for a cube */
  latchCommand=1;
  numIDs=0;
  
  return LATCH_CMD;
}
{NAMES_CMD} {
  BEGIN(0);  /* don't mistake ID for a cube */
  return NAMES_CMD;
}
{END_CMD} return END_CMD;
{NEW_LINE} {
  BEGIN(CUBE_SEARCH);
  return NEW_LINE;
}
<CUBE_SEARCH>{CUBE} yylval.str = yytext; return CUBE;
<LATCH_OPTIONALS>{LATCH_TYPE} yylval.str = yytext; BEGIN(0); return LATCH_TYPE;
<LATCH_OPTIONALS>{LATCH_VAL} yylval.str = yytext; BEGIN(0); return LATCH_VAL;
{ID} {
  yylval.str = yytext; 
  if (latchCommand){
    numIDs++;
    if (numIDs==2){
      BEGIN(LATCH_OPTIONALS);
    }
  }
  return ID;
}
"#"[^\n]* /* eat up comments */
[ \t]+ /* eat up whitespace */
"\\\n" /* eat up line extenders */

%%
