/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef DECOMP_H
#define DECOMP_H

#include "fbddutil.h"
#include "fbddopt.h"

/********************************
  Decomposition Transformation
*********************************/
struct _DecompTransform
{
  DdManager *ddManager;
  struct _List lDTGs; // list of DecompTransformGates.
  char *dType;
};
DecompTransform* dtConstruct(DdManager *ddManager);
void             dtDestruct(DecompTransform *dt);
void             dtAddGate(DecompTransform *dt, 
			   int *newIndexToOld, DdNode *bdd, int numVars,
			   int nodeCount, char reduceRequired);
void             dtAddMuxInput(Model *model, DecompTransform *dt,
			       DdNode *bdd);
int              dtCostCmp(BddProperties *bpOrig, 
			   DecompTransform *dt1, DecompTransform *dt2);
char             dtCostValid(BddProperties *bpOrig, DecompTransform *dt);
void             dtCommit(DecompManager dm, BddClass *bddc);

/********************************
        Bdd Class Heap
*********************************/
BchManager  bchNewManager(Model *model);
void        bchDelManager(BchManager bchm);
void        bchPush(BchManager bchm, BddClass *bddc);
BddClass*   bchPop(BchManager bchm);
int         bchEmpty(BchManager bchm);
char        bchIsActive(BchManager bchm);
void        bchRemove(BchManager bchm, BddClass *bddc);

#endif
