/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef FBDDOPT_H
#define FBDDOPT_H

#include "stdlib.h"
#include "time.h"
#include "fbdd.h"
#include "fbddutil.h"

// COM_DECLARE_CLASS( LocalManager );
COM_DECLARE_CLASS( IFbddCallBack );
COM_DECLARE_CLASS( DecompManager );
COM_DECLARE_CLASS( IDecompCallBack );
COM_DECLARE_CLASS( SeManager );
COM_DECLARE_CLASS( ISeCallBack );
COM_DECLARE_CLASS( ElimManager );
COM_DECLARE_CLASS( SweepManager );
COM_DECLARE_CLASS( ISweepCallBack );
COM_DECLARE_CLASS( BchManager );

typedef enum {
  FBDD_ELIMMODE_CHILD_SINGLE_SUPPORT, 
  FBDD_ELIMMODE_WITH_SS, 
  FBDD_ELIMMODE_AGGRESSIVE,
  FBDD_ELIMMODE_TOTAL
} ElimMode;

/*
 *

	static char*		descRuntime[] = {
		"t%decompose%.2s",
		"t%elim%.2s",
		"t%bddcopy%.2s",
		"t%reduceheap%.2s",
		"t%comsearch%.2s",
		NULL
		};

	ssConstruct(
		fbddGetRuntime(gm), fbddGetMem(gm), 1, descRuntime
		);
	ssBeginTime( fbddGetRuntime(gm) );
	ssBeginTime( fbddGetRuntime(gm) );
	ssEndTime( fbddGetRuntime(gm), 0, FBDD_RUNTIME_DECOMPOSE );
	ssEndTime( fbddGetRuntime(gm), 0, FBDD_RUNTIME_TOTAL );
 
typedef enum {
	FBDD_RUNTIME_ELIM,
	FBDD_RUNTIME_SWEEP,
	FBDD_RUNTIME_DECOMP,
	FBDD_RUNTIME_SIMPLIFY,
	FBDD_RUNTIME_SHARE_EXTRACT,
} FbddRuntimeKind;

struct _LocalMananger {
	FbddManager			global;
	};

#define fbddGetGlobal(lm)		(((LocalManager)lm)->global)
#define fbddGetGlobalMem(lm)	(fbddGetGlobal(lm)->mem)
 */

#ifdef __cplusplus
extern "C" {
#endif

/************************************
	Call Back Interfaces
************************************/

  struct _IFbddCallBack{
    void *objCallee;
    void (*onNewBddc)( void *obj, BddClass *bddc );
    void (*onDelBddc)( void *obj, BddClass *bddc );
  };

  struct _ISweepCallBack{
    void *objCallee;
    void (*beforeSweepBddi)( void *obj, BddInstance *bddi );
    void (*afterSweepBddi)( void *obj, BddInstance *bddi );
  };
  
  struct _ISeCallBack{
    void *objCallee;
    void (*onNewBddc)( void *obj, BddClass *bddc );
    void (*onDelBddc)( void *obj, BddClass *bddc );
  };

  struct _IDecompCallBack{
    void *objCallee;
    void (*beforeRemoveBddi)( void *obj, BddInstance *bddi );
    void (*afterAddBddi)( void *obj, BddInstance *bddi );
    void (*afterDecomposeBddc)( void *obj, void *data );
  };


  /***********************************
           FBDD Manager
  ************************************/
  struct _FbddManager {
    IMem mem;
    
    struct _List lCBs;
    
    SweepManager sm;
    DecompManager dm;
    ElimManager em;
    SeManager sem;
    Model *model;
    
    // User command line options
    int options[NUM_OPTIONS];
  };
  
  void         fbddCBAssign(IFbddCallBack cb, void *objCallee,
			    void (*onNewBddc)( void *obj, BddClass *bddc ),
			    void (*onDelBddc)( void *obj, BddClass *bddc ));
  extern FbddManager _fm;

/*************************************
		 Decomposition
************************************/
struct _DecompManager
{
  FbddManager fm;

  struct _List lCBs;

  struct _IFbddCallBack cbFbdd;
  ListCell cFbddCallBack;

  struct _List lDecompAlgs; // list of decomposition algorithms.
  BchManager bchm;

  // Used by Total Mux Decompose
  struct _HashTable htBddToBddi;
  BddInstance **oldInputVars;
  BddInstance *bddiOrig;
  DdNode *bddTopNode;
  struct _List lSweepBddis;
};

  extern DecompManager decompNewManager(FbddManager gm);
  extern void          decompDelManager(DecompManager lm);
  extern void          decompRun(DecompManager lm);
  extern void          decompDump();
  extern clock_t       decompGetTime();
  extern void          decompAddBddc(DecompManager dm, BddClass *bddc);
  extern void          decompRemBddc(DecompManager dm, BddClass *bddc);

/*************************************
         Sharing Extraction
*************************************/
  struct _SeManager
  {
    FbddManager fm;
    SeMode seMode;
    
    struct _List lCBs;
    
    /* Call Back Instances */
    struct _ISweepCallBack cbSweep;
    ListCell cSweepCallBack;
    struct _IDecompCallBack cbDecomp;
    ListCell cDecompCallBack;
    struct _IFbddCallBack cbFbdd;
    ListCell cFbddCallBack;

    // extractors with less than 'lbMatch' matches are not processed.
    int lbMatch;
    
    // SEExtractors lookup table.
    struct _HashTable htSEEs[5];
    
    // SEExtractors sorted by cost.
    struct _Heap heapSEEs[5-1];
    
    // SEExtractors grouped in lists by cost.
    struct _HashTable htSEELists[5-1];
    
    // TRUE during seRun()
    char seActive;

    /* Sharing Extraction Remainder Cache */
    struct _HashTable htRemainderCache;

    // used by seAddBddiExtractors()
    BddInstance *bddiShareAddExtractors; 
  };
  
  extern SeManager  seNewManager(FbddManager fm, SeMode seMode);
  extern void       seDelManager(SeManager sem);
  extern void       seRun(SeManager lm, int lbMatch);
  extern void       seDump();
  extern void       bddcFreeSeelList(SeManager sem, BddClass *bddc);
  extern void       bddiFreeSeeList(BddInstance *bddi);
  extern void       seAddBddiExtractors(SeManager sem, BddInstance *bddi);
  extern void       seRemBddiExtractors(SeManager sem, BddInstance *bddi);

  extern clock_t    seGetTime();
  extern void       seCBAssign(ISeCallBack cb,
			       void *objCallee,
			       void (*onNewBddc)( void *obj, BddClass *bddc ),
			       void (*onDelBddc)( void *obj, BddClass *bddc ));

/*************************************
             Elimination
*************************************/
  extern ElimManager 	elimNewManager(FbddManager fm, ElimMode mode);
  extern void        	elimDelManager(ElimManager em);
  extern char        	elimRun(ElimManager em, 
				int maxBddNodeChange, 
				clock_t timeLimit);
        extern void        	elimDump();

/*************************************
               Sweep
*************************************/
struct _SweepManager
{
  FbddManager fm;
  DdManager *ddm;

  // List of call backs.
  struct _List lCBs;

  // used by sweepInc
  List lSweepIncBddis;
  IMem memSweepInc;
};

extern SweepManager sweepNewManager(FbddManager fm);
extern void         sweepDelManager(SweepManager sm);
extern BddClass*    sweepBddc(SweepManager sm,
			      BddClass *bddc, 
			      InputConfig inputSimplify, char reorder,
			      int *oldIndexToNew);
extern int          sweepBddi(SweepManager sm, BddInstance *bddi, 
				List lDanglingOutputs);
extern void         sweepDanglingNetwork(SweepManager sm, BddInstance *bddi);
extern void         sweepDanglingLatchNetworks(SweepManager sm);
extern void         sweepIncr(SweepManager sm, IMem mem, List lBddis);
extern void         sweepRun(SweepManager sm);
extern void         sweepCBAssign(ISweepCallBack cb,
                      void *objCallee,
	              void (*beforeSweepBddi)( void *obj, BddInstance *bddi ),
                      void (*afterSweepBddi)( void *obj, BddInstance *bddi ));
       
/*************************************
               Simplify
*************************************/
        extern void         fbddSimplify(FbddManager fm);

#ifdef __cplusplus
}
#endif
		
#endif


