/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#ifndef FBDDUTIL_H
#define FBDDUTIL_H

#include "stdlib.h"
#include "string.h"
#include "metautil.h"
#include "cuddInt.h"
#include "cudd.h"
#include "time.h"

// #define DEBUG_COMPILE

#define CUDD_REORDER_TYPE CUDD_REORDER_SYMM_SIFT
#define BDS_BDD_STARTUP_MEMORY 10*1024*1024
#define BDS_BDD_STARTUP_CACHE  10*1024

#define MAX_BDD_VARS 230
#define MAX_BDD_SIZE 170
#define IDEAL_MAX_BDD_VARS 200

typedef enum{ET_AND, ET_OR, ET_AND01, ET_AND10, ET_XOR, ET_NONE} extractType;

#define PRIME1			12582917
#define PRIME2			4256249
#define PRIME3			741457
#define PRIME4			1618033999
#define PRIME5			551451161
#define PRIME6			517450117

#define CONST_ONE (MAX_BDD_VARS+1)
#define CONST_ZERO (MAX_BDD_VARS+2)

#define THEN_EDGE 1
#define ELSE_EDGE 2

/* Cut Type */
#define CT_NONE 0
#define CT_REPLACE_ONE_EDGES 1
#define CT_REPLACE_ZERO_EDGES 2
#define CT_REPLACE_DANGLING_EDGES 3
#define VAL_REGULAR 1
#define VAL_COMPLEMENT 2
enum {CT_REGULAR, CT_NESTED_BDD, CT_NESTED_BDD_REMAINDER};

// Command line options
#define NUM_OPTIONS 50
typedef enum 
{
  FBDD_OPT_ELIM_DISABLE, 
  FBDD_OPT_DECOMP_DISABLE, 
  FBDD_OPT_SE_DISABLE, 
  FBDD_OPT_MAXEXTSIZE,
  FBDD_OPT_SE_MODE,
  opt_nf, 
  opt_ti, 
  opt_bm, 
  opt_nsweep, 
  opt_rd, 
  opt_nse2, 
  opt_fast, 
  opt_gblelim, 
  opt_printdata, 
  opt_tc, 
  opt_2se, 
  opt_verb,
  opt_conj_ext
} FbddOptions;

typedef struct _BddInstance BddInstance;
typedef struct _BddClass BddClass;
typedef struct _Model Model;
typedef struct _Signal Signal;
typedef struct _DecompTransform  DecompTransform;
typedef int*   InputConfig;
COM_DECLARE_CLASS(BclManager);

// BddClass Gate Type
typedef enum{PRIMARY_INPUT, PRIMARY_OUTPUT, INTERNAL_VARIABLE, 
	       LATCH, CLOCK} bddcType;
struct _BddClass
{
  DdNode* bdd;         // logic function.
  char *name;
  bddcType type;       /* Primary Input, Primary Output, Latch or Internal 
			  signal. */
  int numVars;         // number of input variables for this logic function.
  int numNodes;        // number of bdd nodes

  struct _List lInsts; // list of BddInstances.
  DecompTransform *dt; // cached decomposition of this function.
  
  List *lSeels;       // list of Extractors

  ListCell cBchBddc;   // ListCell in Bch that contains this BddClass.

  unsigned int mark;   // 
                       // 0: modelBddcList()
};

struct _BddInstance
{
  BddClass *bddClass;

  // Used when Class type is internal node
  // index into inputVars is the index of the node
  // ie. node with index i represents BddInstance at inputVars[i]
  struct _BddInstance **inputVars;
  struct _List lFanout;
  ListCell cBddcInst;

  unsigned int mark;
  // bit 0: for mpiNumSS
  // bit 1: for BddiAttachment PARENT
  // bit 2: for BddiAttachment CHILD
  // bit 3: for checkSupportDisjoint
  // bit 4: for spConstructSharedManager && tweAddExtractorsFromBddi
  // bit 5: for equivSharing Bddi aborted
  // bit 6: for twoLiteralExtract var1
  // bit 7: for twoLiteralExtract var2
  // bit 8: latch init value
  // bit 9: latch init value
  // bit 10: sweep ic compute
  // bit 11: bddi sweeped
  // bit 12: modelBddiList/modelBddcList
  // bit 13: twoLiteralExtract : unalign and tweExtractFromBddi
  // bit 14: tweRemDepExtractorsUsingBddiShare
  // bit 15: sweepDanglingLatchNetworks
  //  ListCell cMFBddi; // used during pmElim()

  // Used when Class type is primary input, or if is primary output
  char *id; // for primary inputs and outputs

  struct _List *lSees; // [MAX_EXTRACTOR_SIZE-1];
};

typedef struct _ManArray ManArray;
struct _ManArray
{
  DdManager **ddManagers;
  DdManager *activeManager;
};

struct _Model
{
  //   IMem arena;

  char name[50];
  DdManager *ddManager;
  struct _List lOutputs; // list of output bddis
  struct _List lInputs; // list of input bddis

  struct _List lBOutputs;

  int nextBddcId;

  BclManager bclm;
  int *oldIndexToNewSF;

  BddClass *bddcMux, *bddcMux2, *bddcAnd, *bddcOr, *bddcXnor, *bddcNot, 
    *bddcSF, *bddcNand, *bddcNor, *bddcXor, *bddcLogicZero, 
    *bddcLogicOne, *bddcLatch;
  
  DdNode *nodeSF, *nodeMUX, *nodeMUX2, *nodeAND, *nodeOR, 
    *nodeXNOR, *nodeNOT, *nodeNAND, *nodeNOR, *nodeXOR, 
    *nodeLogicOne, *nodeLogicZero, *nodeAND01, *nodeAND10;

  BddClass *bddcLatchFe,  *bddcLatchRe, *bddcLatchAh, 
    *bddcLatchAl, *bddcLatchAs;

  BddInstance *bddiLogicOne, *bddiLogicZero;

  BddClass *bddcPrimaryOut, *bddcPrimaryIn;
  BddClass *bddcClock;
  BddInstance *bddiDummy;
  BddInstance *bddiGenericClk;

  ManArray manArrayFold;
  ManArray manArrayDecompose;
};

typedef struct _BddProperties BddProperties;
struct _BddProperties
{
  int nodeCount_withComplement;
  int nodeCount_noComplement;
  int numVars;
};

typedef struct _ShareCost ShareCost;
struct _ShareCost
{
  DdNode *shareBdd;
  int nNodeChange;
  int nVarChange;
};

typedef struct _ObjFuncPtr* ObjFuncPtr;
struct _ObjFuncPtr
{
  void *obj;
  void (*func)(void *obj, void *data);
};

/********************************
        BddClass functions 
*********************************/
BddClass*    bddcCreate(char *name, DdNode *bdd, bddcType type, int numVars);
BddInstance* bddcNewInstance(BddClass *bddc);
void         bddcAddInstance(BddClass *bddc, BddInstance* bddi);
void         bddcDestroyKeepInstances(BddClass* bddc);
void         bddcDestruct(BddClass* bddc);
BddClass*    bddcCreateForBdd(Model *model, DdNode *node, 
			      int numVars, int *nextId);
#define      bddcNumInsts(bddc) (listGetSize(&bddc->lInsts))
#define      bddcGetMark(bddc, bit) (bddc->mark&(0x1<<bit))
#define      bddcSetMark(bddc, bit, val) \
  (bddc->mark = (bddc->mark & (0xffff-(0x1<<bit))) | ((val?1:0)<<bit))

/********************************
       BddInstance functions
*********************************/
void         bddiDestruct(BddInstance *bddi);
BddInstance* bddiCopyConstruct(BddInstance *bddi);
void         bddiCollapseGates(IMem mem, BddInstance *bddi, 
			       BddClass *bddcCollapse);
int          bddiReachable(Model *model, BddInstance *bddi);
DdNode*      bddiGetSupportConfig(DdManager *ddManager, 
				  BddInstance *bddiParent, 
				  BddInstance *bddiChild, int *shareSize);
int          bddiGetPosChildInParent(BddInstance *bddiParent, 
				     BddInstance *bddiChild);
#define      bddiNumFanout(bddi) (listGetSize(&bddi->lFanout))
#define      bddiGetMark(bddi, bit) (bddi->mark&(0x1<<bit))
#define      bddiSetMark(bddi, bit, val) \
  (bddi->mark = (bddi->mark & (0xffff-(0x1<<bit))) | ((val?1:0)<<bit))
void printBddi(DdManager *manager, BddInstance *bddi);

/********************************
         Model functions
*********************************/
Model*       modelCreate(char *name);
void         modelDestroy(Model *model);
void         modelBddcList(IMem mem, Model *model, List lBddcs);
void         modelBddiList(IMem mem, Model *model, List list);
void         modelAllBddiList(IMem mem, Model *model, List list);
int          modelNumGates(Model *model);
int          modelLitCount(Model *model);
void         modelGetCost(Model *model, int *nc, 
			  int *bestCase, int *worstCase);
int          modelGetNodeCount(Model *model);
void         modelBackup(Model *model);
void         modelClearBackup(Model *model);
void         modelSwapBackup(Model *model);
void         modelRestore(Model *model);
#define      modelHasBackup(model) (listGetSize(&model->lBOutputs)>0)
int          modelCountRepeatingNOTs(Model *model);
void         modelPrintGates(Model *model);
void         modelComputeFanout(Model *model);

/********************************
       DdManager functions
*********************************/
DdManager*   newDdManager();
void         manArrayConstruct(ManArray *manArray);
void         manArrayDestruct(ManArray *manArray);
DdManager*   getDdManager(ManArray *manArray, int numVars);
void         releaseDdManager(ManArray *manArray);

/********************************
         Node Mark Lookup
*********************************/
typedef struct _Nml{
  IMem arena;
  HashTable hashTable;
} Nml;

void     nmlConstruct(Nml *nml);
void     nmlDestruct(Nml *nml);
DdNode*  nmlGetNode(Nml *nml, DdNode *node);
int      nmlGetMark(Nml *nml, DdNode *node);
void     nmlPutNode(Nml *nml, DdNode *node, DdNode *nodeB);
void     nmlPutMark(Nml *nml, DdNode *node, int marking);
int      nmlGetBddiMark(Nml *nml, BddInstance *bddi);
void     nmlPutBddiMark(Nml *nml, BddInstance *bddi, int marking);
void     nmlRemove(Nml *nml, DdNode *node);
void     nmlIncrMark(Nml *nml, DdNode *node);
long     nmlGetLMark(Nml *nml, DdNode *node);
void     nmlPutLMark(Nml *nml, DdNode *node, long marking);

/********************************
        Level First Search
*********************************/
typedef struct _LevelFirstSearch{
  IMem            arena;
  struct _Heap    heap;
  Nml             nml;
  DdManager*      ddManager;
} LevelFirstSearch;

void      lfsConstruct(LevelFirstSearch *lfs, DdManager *ddManager,
		       DdNode *node);
void      lfsDestruct(LevelFirstSearch *lfs);
DdNode*   lfsNextNode(LevelFirstSearch *lfs);

/********************************
         Bdd Class Lookup 
*********************************/
BclManager  bclNewManager();
void        bclDelManager(BclManager bcl);
BddClass*   bclGet(BclManager bcl, DdNode *node);
void        bclPut(BclManager bcl, DdNode *node, BddClass *bddClass);
void        bclRemove(BclManager bcl, DdNode *node);
void        bclRemoveBddi(BclManager bcl, Model *model, BddInstance *bddi);
void        populateBclWithBasicGates(BclManager bcl, Model *model);

/********************************
           InputConfig 
*********************************/
InputConfig icCompute(IMem mem, int numVars, char findConsts, 
		      BddInstance **inputVars);
int*        icComputeStringKey(IMem mem, int numVars, 
			       InputConfig inputConfig);
char        icCompare(int numVars, InputConfig ic1, InputConfig ic2);
int         icNumSS(InputConfig ic);
#define     icCanSimplify(ic) (ic[0]>0)
#define     icReadVar(ic, v) (ic[v+1])
#define     icFree(mem, ic) (comFree(mem, ic))

/********************************
          Basic Gates
*********************************/
DdNode* buildLogicZero(DdManager *manager);
DdNode* buildLogicOne(DdManager *manager);
DdNode* buildSF(DdManager *manager);
DdNode* buildNOT(DdManager *manager);
DdNode* buildForwarding(DdManager *manager);
DdNode* buildMUX(DdManager *manager);
DdNode* buildMUX2(DdManager *manager);
DdNode* buildOR(DdManager *manager);
DdNode* buildOROfSize(DdManager *ddManager, int size);
DdNode* buildAND(DdManager *manager);
DdNode* buildAND01(DdManager *manager);
DdNode* buildAND10(DdManager *manager);
DdNode* buildXNOR(DdManager *manager);
#define isBasicGate(model, bddc) \
   (bddc==model->bddcAnd || bddc==model->bddcOr \
    || bddc==model->bddcMux || bddc==model->bddcMux2 \
    || bddc==model->bddcXnor || bddc==model->bddcNot \
    || bddc==model->bddcSF || bddc==model->bddcNand \
    || bddc==model->bddcNor || bddc==model->bddcXor \
    || bddc==model->bddcLogicOne || bddc==model->bddcLogicZero \
    || bddc->type!=INTERNAL_VARIABLE)
#define isBasicGateBdd(model, bdd1) \
    (bdd1==model->bddcAnd->bdd || bdd1==model->bddcOr->bdd \
     || bdd1==model->bddcMux->bdd || bdd1==model->bddcMux2->bdd \
     || bdd1==model->bddcXnor->bdd || bdd1==model->bddcNot->bdd \
     || bdd1==model->bddcSF->bdd || bdd1==model->bddcNand->bdd \
     || bdd1==model->bddcNor->bdd || bdd1==model->bddcXor->bdd \
     || bdd1==model->bddcLogicOne->bdd \
     || bdd1==model->bddcLogicZero->bdd)

/********************************
     Apply Transformations
*********************************/
void         updateBddisForRoot(Model *model, BddInstance *bddi, 
				BddClass *bddc,
				int *oldIndexToNewA, int arraySizeA,
				BddInstance **oldInputVars,
				int *oldIndexToNewA2, int arraySizeA2,
				BddInstance **oldInputVars2,
				int *oldIndexToNewB, int arraySizeB,
				int additionalPos, 
				BddInstance *additionalBddi);
BddInstance* createBddiForExt(Model *model, BddClass *bddc, 
			      BddInstance **oldInputVars, int oldNumVars,
			      int *oldIndexToNewA, int arraySizeA, 
			      BddInstance **oldInputVars2, int oldNumVars2,
			      int *oldIndexToNewA2, int arraySizeA2, 
			      int *oldIndexToNewB, int arraySizeB,
			      BddInstance *bddiFanout1, 
			      BddInstance *bddiFanout2);

/********************************
        BDD Manipulation
*********************************/
DdNode*     foldBddClass(Model *model, DdNode *topNode, 
	          int *NumVars, int **oldIndexToNew, int *arraySize,
                  DdNode *nReplace, DdNode *nReplaceWith);
DdNode*     foldBddUpdateOldIndexToNew(Model *model, DdNode *bdd, 
		  int **oldIndexToNew, int arraySize);
DdNode*     bddCopyAndReIndex(DdManager *srcManager, DdNode *srcNode, 
		  DdManager *dstManager, 
		  int *numVars, int **oldIndexToNew, int *arraySize,
		  DdNode *nReplace, DdNode *nReplaceWith);
DdNode *    bddCopyAndReIndexWithCut(DdManager *srcManager, DdNode *srcNode, 
		  DdManager *dstManager, 
                  int *numVars, int **oldIndexToNew, int *arraySize,
                  char cutType, DdNode *replaceWith, int cutLevel);
DdNode*     bddCopyAndReIndexNestedBdd(DdManager *srcManager, DdNode *srcNode, 
		  DdManager *dstManager, 
		  int *numVars, int **oldIndexToNew, int *arraySize,
		  int copyType, 
		  DdNode *nReplaceRoot, 
		  DdNode *nReplaceThen, DdNode *nReplaceElse,
		  int *nestedBddIndex);
DdNode*     bddCopy(DdManager* srcManager, 
                  DdNode *srcNode, DdManager *dstManager, 
		  int *oldIndexToNew, DdNode *nReplace, DdNode *nReplaceWith);
DdNode*     bddComputeOperation(DdManager *ddManager, extractType et, 
				DdNode *n1, DdNode *n2);
#define     getComplementAdjustedNode(nChild, nParent) \
     (Cudd_IsComplement(nChild) == Cudd_IsComplement(nParent))\
     ?Cudd_Regular(nChild):Cudd_Complement(nChild)

void        printBdd(DdNode *node);
void        bddGetProperties(DdNode *node, BddProperties *bp);
int         bddCmpCost(BddProperties *bpOrig,
	       BddProperties *bpA1, BddProperties *bpB1, BddProperties *bpC1, 
	       BddProperties *bpA2, BddProperties *bpB2, BddProperties *bpC2);

/********************************
  HashSchemes Support Functions 
*********************************/
COMDECL  _hashQuery(IHashScheme si, const ComIID iid, IGeneric *i);
COMDECLU _hashAddref( IHashScheme si );
COMDECLU _hashRelease( IHashScheme si );

/********************************
         Common Utilities
*********************************/
int getOldIndexFromNew(int *oldIndexToNew, int arraySize, int newIndex);
void diPrintInfo(char *sFile, Model *model);

/********************************
         Debug Functions 
*********************************/
int checkFaninFanout(Model *model);
int checkBddcsBddis(Model *model);
int checkSimplified(Model *model);
int checkBddi(BddInstance *bddi);
int checkBddiReachable(Model *model, BddInstance *bddi);

/********************************
        Call Back Routines
*********************************/
#define cbRun(ICallBackType, objCaller, func, data) \
        {\
          ListCell listCell; \
          listForeach(&((objCaller)->lCBs), listCell) {\
            ICallBackType cb = listData(ICallBackType, listCell); \
            if ((cb)->func != NULL) \
              (cb)->func(((cb)->objCallee), data);\
          }\
	}
#define cbAdd(obj, listCell, callBack) \
        {\
          listNewCell(_fbddHeap, listCell, (void*)callBack); \
          listAppendElement(&((obj)->lCBs), listCell); \
	}
#define cbRemove(obj, listCell) \
          listDelElement(&((obj)->lCBs), listCell);
          

/********************************
    Input / Output Functions
*********************************/
void    modelWriteDot(Model *model, char *instFile, char *bddcFile);
void    bddWriteDot(char *filename, DdManager *manager, DdNode *bdd);
void    bddWriteDDCal(char *filename, DdManager *manager, DdNode *bdd);

/********************************
             Other
*********************************/
int        scCmp(ShareCost *sc1, ShareCost *sc2);
void       printArray(int *array, int numElements);
void       mergeOldIndexToNew(int *oldIndexToNew1, int arraySize1, 
			      int *oldIndexToNew2, int arraySize2);
extern long util_cpu_time (void);

extern int   maxExtSize;
extern IMem  _fbddHeap;

extern clock_t tTotal;
extern clock_t tSimplify;
extern clock_t tSweep;
extern clock_t tElim;
extern clock_t tDecomp;
extern clock_t tSE;

extern clock_t tBddCopy;
extern clock_t tReduceHeap;
extern clock_t tFoldBddClass;
extern clock_t tCollapse;

#endif
