/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <assert.h>
#include "metautil.h"
#include "fbddutil.h"

void
lfsConstruct(LevelFirstSearch *lfs, DdManager *ddManager, DdNode *node)
{
  HeapCell hc;

  lfs->arena = _fbddHeap;
  heapConstruct( &lfs->heap, lfs->arena, 100);
  nmlConstruct(&lfs->nml);
  lfs->ddManager = ddManager;
  hc = heapNewCell( &lfs->heap, 
		    Cudd_ReadPerm(ddManager, Cudd_NodeReadIndex(node)), 
		    Cudd_Regular(node) );
  heapInsertElement( &lfs->heap, hc );
}

void
lfsDestruct(LevelFirstSearch *lfs)
{
  nmlDestruct(&lfs->nml);
  while(!heapIsEmpty(&lfs->heap)){
    HeapCell hc = heapDelMin(&lfs->heap);
    comFree(heapMem(&lfs->heap), hc);
  }
  heapDelAll(&lfs->heap);
}

DdNode *
lfsNextNode(LevelFirstSearch *lfs)
{
  HeapCell hc;
  DdNode *node;
  DdNode *nNew;

  if (heapIsEmpty(&lfs->heap))
    return NULL;

  hc = heapDelMin(&lfs->heap);
  node = heapData(DdNode*, hc);
  comFree(lfs->arena, hc);

  if (Cudd_IsConstant(node))
    return node;

  // add the children of 'node' to the heap
  nNew = Cudd_T(node);
  if (nmlGetMark(&lfs->nml, Cudd_Regular(nNew)) == 0){
    nmlPutMark(&lfs->nml, Cudd_Regular(nNew),1);
    hc = heapNewCell( &lfs->heap, 
		      Cudd_ReadPerm(lfs->ddManager, Cudd_NodeReadIndex(nNew)), 
		      Cudd_Regular(nNew) );
    heapInsertElement( &lfs->heap, hc );
  }
  
  nNew = Cudd_E(node);
  if (nmlGetMark(&lfs->nml, Cudd_Regular(nNew)) == 0){
    nmlPutMark(&lfs->nml, Cudd_Regular(nNew) ,1);
    hc = heapNewCell( &lfs->heap, 
		      Cudd_ReadPerm(lfs->ddManager, Cudd_NodeReadIndex(nNew)), 
		      Cudd_Regular(nNew) );
    heapInsertElement( &lfs->heap, hc );
  }
  return node;
}

