/*
 * Copyright (c) 2005 The University of Toronto and the Synthesis Group.
 * All rights reserved.
 * 
 * This file is part of the FBDD.  The FBDD is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Toronto at jzhu@eecg.toronto.edu or +1-416-946-5971.
 * 
 * The FBDD is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the FBDD; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <assert.h>

#include "fbddutil.h"

DdNode *
bddExtract(DdManager *ddManager, DdNode *bddOrig, DdNode *bddExtract,
	   int extractIndex)
{
  DdNode *bddRemainder;
  DdNode *nX;
  DdNode *bddCare;
  DdNode *bddComplete;

  // create new node used to represent bddExtract.
  // Assume bddOrig is in the global manager (the top most node has the 
  // largest index).
  //  nX = Cudd_bddIthVar(ddManager, extractIndex);
  nX = Cudd_ReadVars(ddManager, extractIndex);
  assert(nX);
  Cudd_Ref(nX);

  // generate care function.
  {
    DdNode *bddXorFG, *bddXnorFX;
    
    bddXorFG = Cudd_bddXor(ddManager, bddOrig, bddExtract);
    Cudd_Ref(bddXorFG);

    bddXnorFX = Cudd_bddXnor(ddManager, bddOrig, nX);    
    Cudd_Ref(bddXnorFX);

    bddCare = Cudd_bddOr(ddManager, bddXorFG, bddXnorFX);
    Cudd_Ref(bddCare);
    Cudd_RecursiveDeref(ddManager, bddXorFG);
    Cudd_RecursiveDeref(ddManager, bddXnorFX);
  }

  // generate completely specified function
  {
    DdNode *bddTmp;

    bddTmp = Cudd_bddOr(ddManager, nX, Cudd_Not(bddExtract));
    Cudd_Ref(bddTmp);
    
    bddComplete = Cudd_bddAnd(ddManager, bddTmp, bddOrig);
    Cudd_Ref(bddComplete);
    Cudd_RecursiveDeref(ddManager, bddTmp);
  }

  bddRemainder = Cudd_bddRestrict(ddManager, bddComplete, bddCare);
  Cudd_Ref(bddRemainder);
  Cudd_RecursiveDeref(ddManager, bddCare);
  Cudd_RecursiveDeref(ddManager, bddComplete);
  Cudd_RecursiveDeref(ddManager, nX);

  return bddRemainder;
}

